# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from airflow import DAG
from airflow.contrib.operators import gcs_to_bq, kubernetes_pod_operator

default_args = {
    "owner": "Google",
    "depends_on_past": False,
    "start_date": "2021-03-01",
}


with DAG(
    dag_id="world_bank_wdi.country_summary",
    default_args=default_args,
    max_active_runs=1,
    schedule_interval="@daily",
    catchup=False,
    default_view="graph",
) as dag:

    # Run CSV transform within kubernetes pod
    country_summary_transform_csv = kubernetes_pod_operator.KubernetesPodOperator(
        task_id="country_summary_transform_csv",
        startup_timeout_seconds=600,
        name="country_summary",
        namespace="composer",
        service_account_name="datasets",
        image_pull_policy="Always",
        image="{{ var.json.world_bank_wdi.container_registry.run_csv_transform_kub }}",
        env_vars={
            "SOURCE_URL": "gs://pdp-feeds-staging/RelayWorldBank/WDI_csv/WDICountry.csv",
            "SOURCE_FILE": "files/data.csv",
            "COLUMN_TO_REMOVE": "Unnamed: 30",
            "TARGET_FILE": "files/data_output.csv",
            "TARGET_GCS_BUCKET": "{{ var.value.composer_bucket }}",
            "TARGET_GCS_PATH": "data/world_bank_wdi/country_summary/data_output.csv",
            "PIPELINE_NAME": "country_summary",
            "CSV_HEADERS": '["country_code","short_name","table_name","long_name","two_alpha_code","currency_unit","special_notes","region","income_group","wb_2_code","national_accounts_base_year","national_accounts_reference_year","sna_price_valuation","lending_category","other_groups","system_of_national_accounts","alternative_conversion_factor","ppp_survey_year","balance_of_payments_manual_in_use","external_debt_reporting_status","system_of_trade","government_accounting_concept","imf_data_dissemination_standard","latest_population_census","latest_household_survey","source_of_most_recent_income_and_expenditure_data","vital_registration_complete","latest_agricultural_census","latest_industrial_data","latest_trade_data","latest_water_withdrawal_data"]',
            "RENAME_MAPPINGS": '{"Country Code":"country_code","Short Name":"short_name","Table Name":"table_name","Long Name":"long_name","2-alpha code":"two_alpha_code","Currency Unit":"currency_unit","Special Notes":"special_notes","Region":"region","Income Group":"income_group","WB-2 code":"wb_2_code","National accounts base year":"national_accounts_base_year","National accounts reference year":"national_accounts_reference_year","SNA price valuation":"sna_price_valuation","Lending category":"lending_category","Other groups":"other_groups","System of National Accounts":"system_of_national_accounts","Alternative conversion factor":"alternative_conversion_factor","PPP survey year":"ppp_survey_year","Balance of Payments Manual in use":"balance_of_payments_manual_in_use","External debt Reporting status":"external_debt_reporting_status","System of trade":"system_of_trade","Government Accounting concept":"government_accounting_concept","IMF data dissemination standard":"imf_data_dissemination_standard","Latest population census":"latest_population_census","Latest household survey":"latest_household_survey","Source of most recent Income and expenditure data":"source_of_most_recent_income_and_expenditure_data","Vital registration complete":"vital_registration_complete","Latest agricultural census":"latest_agricultural_census","Latest industrial data":"latest_industrial_data","Latest trade data":"latest_trade_data"}',
        },
        resources={"request_memory": "2G", "request_cpu": "1"},
    )

    # Task to load CSV data to a BigQuery table
    load_country_summary_to_bq = gcs_to_bq.GoogleCloudStorageToBigQueryOperator(
        task_id="load_country_summary_to_bq",
        bucket="{{ var.value.composer_bucket }}",
        source_objects=["data/world_bank_wdi/country_summary/data_output.csv"],
        source_format="CSV",
        destination_project_dataset_table="world_bank_wdi.country_summary",
        skip_leading_rows=1,
        allow_quoted_newlines=True,
        write_disposition="WRITE_TRUNCATE",
        schema_fields=[
            {
                "name": "country_code",
                "type": "string",
                "description": "A short alphabetic or numeric geographical code representing a specific country or area. In April 2016 WDI has adopted the three digit ISO code to identify countries in its databases.",
                "mode": "nullable",
            },
            {
                "name": "short_name",
                "type": "string",
                "description": "Country name the country's official name as approved by the relevant body (Italy is used as an example): conventional short form (Italy).",
                "mode": "nullable",
            },
            {
                "name": "table_name",
                "type": "string",
                "description": "Country name the country's name as approved by the relevant body (Italy is used as an example): conventional short form (Republic of Italy).",
                "mode": "nullable",
            },
            {
                "name": "long_name",
                "type": "string",
                "description": "Country name the country's name as approved by the relevant body (Italy is used as an example): conventional long form (Republic of Italy).",
                "mode": "nullable",
            },
            {
                "name": "two_alpha_code",
                "type": "string",
                "description": "ISO-2 digit alpha code value of the country. Re: www.iso.org",
                "mode": "nullable",
            },
            {
                "name": "currency_unit",
                "type": "string",
                "description": "Currency unit (such as the dollar euro peso rupee) issued as a coin or banknote a standard unit of value and a unit of account.",
                "mode": "nullable",
            },
            {
                "name": "special_notes",
                "type": "string",
                "description": "Country special notes for data and metadata users.",
                "mode": "nullable",
            },
            {
                "name": "region",
                "type": "string",
                "description": "World Bank region to which a country belongs.",
                "mode": "nullable",
            },
            {
                "name": "income_group",
                "type": "string",
                "description": "World Bank income group to which a country belongs after the country classification is completed every year.",
                "mode": "nullable",
            },
            {
                "name": "wb_2_code",
                "type": "string",
                "description": "World Bank 2 digit code value of the country.",
                "mode": "nullable",
            },
            {
                "name": "national_accounts_base_year",
                "type": "string",
                "description": "Base year is the base or pricing period used for constant price calculations in the country’s national accounts. Price indexes derived from national accounts aggregates such as the implicit deflator for gross domestic product (GDP) express the price level relative to base year prices.",
                "mode": "nullable",
            },
            {
                "name": "national_accounts_reference_year",
                "type": "string",
                "description": "Reference year is the year in which the local currency constant price series of a country is valued. The reference year is usually the same as the base year used to report the constant price series. However when the constant price data are chain linked the base year is changed annually so the data are rescaled to a specific reference year to provide a consistent time series. When the country has not rescaled following a change in base year World Bank staff rescale the data to maintain a longer historical series. To allow for cross-country comparison and data aggregation constant price data reported in World Development Indicators are rescaled to a common reference year (2010) and currency (U.S. dollars).",
                "mode": "nullable",
            },
            {
                "name": "sna_price_valuation",
                "type": "string",
                "description": "SNA price valuation shows whether value added in the national accounts is reported at basic prices (B) or producer prices (P). Producer prices include taxes paid by producers and thus tend to overstate the actual value added in production. However value added can be higher at basic prices than at producer prices in countries with high agricultural subsidies.",
                "mode": "nullable",
            },
            {
                "name": "lending_category",
                "type": "string",
                "description": "World Bank lending category to which a country belongs after the operational guidelines are completed every year.",
                "mode": "nullable",
            },
            {
                "name": "other_groups",
                "type": "string",
                "description": "Other types of grouping approved by the WDI team.",
                "mode": "nullable",
            },
            {
                "name": "system_of_national_accounts",
                "type": "string",
                "description": "System of National Accounts identifies whether a country uses the 1968 1993 or 2008 System of National Accounts (SNA). The 2008 SNA is an update of the 1993 SNA and retains its basic theoretical framework.",
                "mode": "nullable",
            },
            {
                "name": "alternative_conversion_factor",
                "type": "string",
                "description": "Alternative conversion factor identifies the countries and years for which a World Bank–estimated conversion factor has been used in place of the official exchange rate (line rf in the International Monetary Fund’s [IMF] International Financial Statistics). See later in Sources and methods for further discussion of alternative conversion factors.",
                "mode": "nullable",
            },
            {
                "name": "ppp_survey_year",
                "type": "string",
                "description": "Purchasing power parity (PPP) survey year is the latest available survey year for the International Comparison Program’s estimates of PPPs.",
                "mode": "nullable",
            },
            {
                "name": "balance_of_payments_manual_in_use",
                "type": "string",
                "description": "Balance of Payments Manual in use refers to the classification system used to compile and report data on balance of payments. 6 refers to the 6th edition of the IMF’s Balance of Payments Manual (2009).",
                "mode": "nullable",
            },
            {
                "name": "external_debt_reporting_status",
                "type": "string",
                "description": "External debt shows debt reporting status for 2016 data. A indicates that data are as reported P that data are based on reported or collected information but include an element of staff estimation and E that data are World Bank staff estimates.",
                "mode": "nullable",
            },
            {
                "name": "system_of_trade",
                "type": "string",
                "description": "System of trade refers to the United Nations general trade system (G) or special trade system (S). Under the general trade system goods entering directly for domestic consumption and goods entered into customs storage are recorded as imports at arrival. Under the special trade system goods are recorded as imports when declared for domestic consumption whether at time of entry or on withdrawal from customs storage.",
                "mode": "nullable",
            },
            {
                "name": "government_accounting_concept",
                "type": "string",
                "description": "Government finance accounting concept is the accounting basis for reporting central government financial data. For most countries government finance data have been consolidated (C) into one set of accounts capturing all central government fiscal activities. Budgetary central government accounts (B) exclude some central government units.",
                "mode": "nullable",
            },
            {
                "name": "imf_data_dissemination_standard",
                "type": "string",
                "description": "IMF data dissemination standard shows the countries that subscribe to the IMF’s Special Data Dissemination Standard (S) or Enhanced General Data Dissemination System (e-GDDS). S refers to countries that subscribe to the SDDS and have posted data on the Dissemination Standards Bulletin Board at http://dsbb.imf.org. S+ countries must observe additional coverage periodicity and timeliness requirements. e-G refers to countries that subscribe to the e-GDDS.",
                "mode": "nullable",
            },
            {
                "name": "latest_population_census",
                "type": "string",
                "description": "Latest population census shows the most recent year in which a census was conducted and in which at least preliminary results have been released. The preliminary results from the very recent censuses could be reflected in timely revisions if basic data are available such as population by age and sex as well as the detailed definition of counting coverage and completeness.",
                "mode": "nullable",
            },
            {
                "name": "latest_household_survey",
                "type": "string",
                "description": "Latest demographic education or health household survey indicates the household surveys used to compile the demographic education and health data. Detailed information on each survey is provided through the following links:  AIDS Indicator Survey Demographic and Health Survey Health Issues Survey and Malaria Indicator Survey - www.dhsprogram.com; Living Standards Measurement Study -  www.worldbank.org/lsms; Multiple Indicator Cluster Survey - http://mics.unicef.org; Reproductive Health Survey -  www.cdc.gov/reproductivehealth; World Health Survey is available at - www.who.int/healthinfo/survey/en.",
                "mode": "nullable",
            },
            {
                "name": "source_of_most_recent_income_and_expenditure_data",
                "type": "string",
                "description": "Source of most recent income and expenditure data shows household surveys that collect income and expenditure data. Names and detailed information on household surveys can be found on the website of the International Household Survey Network (www.surveynetwork.org). Core Welfare Indicator Questionnaire Surveys (CWIQ) developed by the World Bank measure changes in key social indicators for different population groups—specifically indicators of access utilization and satisfaction with core social and economic services. Expenditure survey/budget surveys (ES/BS) collect detailed information on household consumption as well as on general demographic social and economic characteristics. Integrated household surveys (IHS) collect detailed information on a wide variety of topics including health education economic activities housing and utilities.",
                "mode": "nullable",
            },
            {
                "name": "vital_registration_complete",
                "type": "string",
                "description": "Vital registration complete identifies countries that report at least 90 percent complete registries of vital (birth and death) statistics to the United Nations Statistics Division and are reported in its Population and Vital Statistics Reports. Countries with complete vital statistics registries may have more accurate and more timely demographic indicators than other countries.",
                "mode": "nullable",
            },
            {
                "name": "latest_agricultural_census",
                "type": "string",
                "description": "Latest agricultural census shows the most recent year in which an agricultural census was conducted or planned to be conducted as reported to the Food and Agriculture Organization.",
                "mode": "nullable",
            },
            {
                "name": "latest_industrial_data",
                "type": "integer",
                "description": "Latest industrial data show the most recent year for which manufacturing value added data at the three-digit level of the International Standard Industrial Classification (revision 2 or 3) are available in the United Nations Industrial Development Organization database.",
                "mode": "nullable",
            },
            {
                "name": "latest_trade_data",
                "type": "integer",
                "description": "Latest trade data show the most recent year for which structure of merchandise trade data from the United Nations Statistics Division’s Commodity Trade (Comtrade) database are available.",
                "mode": "nullable",
            },
            {
                "name": "latest_water_withdrawal_data",
                "type": "string",
                "description": "Latest water withdrawal data show the most recent year for which data on freshwater withdrawals have been compiled from a variety of sources.",
                "mode": "nullable",
            },
        ],
    )

    country_summary_transform_csv >> load_country_summary_to_bq
